/*****************************************************************************

  CEdge.h

  Stephen Freund
  freunds@cs.stanford.edu

  This class maintains a single, directed edge in the graph.  It has
  queues on both ends for each commodity.  Any adjustment of these
  queues should go through the edge in order to make sure that the
  heaps are updated correctly if necessary.  

  Each Edge is also able to balance itself, using the process in
  Algorithm 2 in the paper.

  To use the Edge, create it and set it up in the graph.  Once the
  number of commodities is known, finishing initalizing the edge
  by calling InitQueues.  This will set up the heap and queues and
  prepare the edge to be used in the graph.

******************************************************************************/


#ifndef _H_CEdge
#define	_H_CEdge

#include <iostream.h>

class CNode;
class CCommodity;
class CHeap;

#include "CEdgeQueue.h"
#include "Parameter.h"

class CEdge {
 private:
  CNode *itsHead;
  CNode *itsTail;
  CHeap *itsDerivativeHeap;
  
  double itsCapacity;          /* starting capacity */
  double itsOriginalCapacity;  /* used as capacity increase over iterations */
  
  int itsNumCommodities;       /* number of commodities in graph */
  CEdgeQueue **itsQueues;      /* array of head & tail queues    */
  
 public:

  /* constructor / destructor */
  CEdge();
  CEdge(CNode *tail, CNode *head, double theCap = 0.0);
  ~CEdge();

  /* initialization */
  void InitQueues(int NumCommodities, CCommodity *comm[]);		
  
  /* accessors */
  CNode *GetHead(void)         { return itsHead; }
  CNode *GetTail(void)         { return itsTail; }
  double GetCapacity(void)     { return itsCapacity;  }
  void SetCapacity(double cap) { itsOriginalCapacity = itsCapacity = cap; }
  double GetOrigCapacity(void)     { return itsOriginalCapacity;  }
  void SetOrigCapacity(double cap) { itsOriginalCapacity = cap; }

  /* operations on individual queues.  pos should be either head
     or tail, and index the number id of the commodity.  */
  double GetQueueSize(EdgePos pos, int index)  
    { return itsQueues[index]->GetCommodityLevel(pos); }
  void SetQueueSize(EdgePos pos, int index, double level);
  void AdjustQueueSize(EdgePos pos, int index, double deltaAmount);

  /* move amount of the commodity with id of index from the tail to
     the head */
  void MoveAcrossEdge(int index, int amount); 

  /* capacity / height adjustments */
  void HalveQueueSizes(void);                  /* halve all queues  */
  void DivideQueueSizes(double a);
  void IncreaseCapacity(void)                  /* increase capacity */
    { itsCapacity += itsOriginalCapacity * GetSigma(); };
  void HalveCapacity(void)	 { itsCapacity /= 2; };
  void DivideCapacity(double amount)  { itsCapacity /= amount; }

  /* Balance each commodity across the edge. */
  void BalanceEdge(void);
  
  /* I/O helpers */
  friend ostream& operator<<(ostream&,const CEdge&);
  void PrintQueues(ostream& out) const;
  void PrintFlow(void);

  double GetFlow(int index)  { return itsQueues[index] -> GetFlow(); }
};

#endif


