/******************************************************************************

  CGraph.C

  Stephen Freund
  freunds@cs.stanford.edu

  Implementation of the CGraph class.

******************************************************************************/

#include "CGraph.h"
#include "CNode.h"
#include "CCommodity.h"
#include "Utility.h"
#include "Parameter.h"
#include <stdio.h>
#include <math.h>


/*
 * Use this when you do not know how many nodes the graph will
 * contain. 
 */
CGraph::CGraph()  {
  itsNumNodes = 0;
  itsNumCommodities = 0;
  itsCommodities = NULL;
  itsNodes = NULL;
}


/*
 * 
 */
CGraph::CGraph(int numberOfNodes)  {
  itsNumNodes = numberOfNodes;
  InitNodes();
  InitCommodities();
}


/*
 * Create a CNode object for each node in the graph.
 */
void CGraph::InitNodes(void) {
  int i;
  itsNodes = new (CNode*)[itsNumNodes];
  for (i=0;i<itsNumNodes;i++)  {
    itsNodes[i] = new CNode(this,i);
  }
}

/*
 * Create the commodities for the graph.
 */
void CGraph::InitCommodities(void) {
  int i;
  itsCommodities = new(CCommodity *)[itsNumCommodities];
  for (i=0;i<itsNumCommodities;i++)  {
    itsCommodities[i] = NULL;
  }
}
	

/*
 * output all the information about a graph.
 */     	
ostream& operator<<(ostream& out,const CGraph& graph) {
  int i;
  out << "Num Nodes: " << graph.GetNumNodes() << "\n";
  
  for (i=0;i<graph.GetNumNodes();i++)  {
    out << *(graph.GetNode(i));
  }
  
  out << "\n";
  for (i=0;i<graph.GetNumCommodities();i++) {
    out << *(graph.GetCommodity(i)) << "\n";
  }
  
  return(out);
}


/*
 * Read in the topology from a stream.  The format of the input should 
 * be the number of nodes, followed by an n x n matrix of numbers, where
 * a x indicates the presence of an edge of capacity scale * x and a 0 
 * indicates no edge.  Since the edges are all bi-directional, only specify 
 * edges in the upper triangle of the matrix.  This function will make edges 
 * going both ways.  
 */ 
void CGraph::ReadTopology(istream& in, double scale)  {
  int u,v;
  CNode *NU, *NV;
  int edge;  

  itsNumEdges = 0;

  for (u=0; u < itsNumNodes; u++)  {
    for (v=0; v < itsNumNodes; v++)  {
      in >> edge;
      if (edge != 0) {
	itsNumEdges+=2;

	NU = itsNodes[u];
	NV = itsNodes[v];
	
	NU -> AddEdge(NV);
	NV -> AddEnteringEdge(NU->GetEdge(NV));
	NU -> GetEdge(NV) -> SetCapacity(edge * scale);
	
	NV -> AddEdge(NU);
	NU -> AddEnteringEdge(NV->GetEdge(NU));
	NV -> GetEdge(NU) -> SetCapacity(edge * scale);
	
      }
    }
  }	
}

/*
 * Read in the commodities from a stream.  The first item on the stream
 * should be the number of commodities.  Then, there should be a matrix.
 * a 0 in the matrix represents no commodity must go from the nodes at the
 * given horizontal and vertical indexes.  A number represents the amount
 * of a commodity which must flow from the row's node to the column's.
 * Again, put values only in the upper triangle.  All commodities will
 * be scaled by scale.
 */
void CGraph::ReadCommodities(istream& in, double scale)  {
  int u,v;
  int demand;
  int count = 0;
  for (u=0; u < itsNumNodes; u++)  {
    for (v=0; v < itsNumNodes; v++)  {
      in >> demand;
      if (demand != 0) {
	itsCommodities[count] = 
	  new CCommodity(count, itsNodes[u], itsNodes[v], demand*scale);
	count++;
	itsCommodities[count] = 
	  new CCommodity(count, itsNodes[v], itsNodes[u], demand*scale);
	count++;
      }
    }
  }	
}


/*
 * Read the data from a stream.
 */
istream& operator>>(istream& in, CGraph& graph) {

  graph.ReadGraphFromMatrixInput(in, CAPACITY, 1);
  return(in);
}

void CGraph::ReadGraphFromMatrixInput(istream& in, double EdgeScale,
				      double CommodityScale) {
  int edgeExists;
  
  in >> itsNumNodes;
  InitNodes();
  ReadTopology(in, EdgeScale);
  
  in >> itsNumCommodities;
  itsNumCommodities *= 2;
  InitCommodities();
  ReadCommodities(in, CommodityScale);
}

void CGraph::ReadGraphFromFormattedInput(istream& in, double EdgeScale,
					 double CommodityScale)  {
  int commandChar;
  int u,v;
  CNode *NU, *NV;
  int val;
  char buf[30];
  int count = 0;

  itsNumEdges = itsNumNodes = itsNumCommodities = 0;

  while (in)  {
    in >> commandChar;
    switch (commandChar)  {
    case 'c':
      while ((commandChar = getchar()) != EOF && commandChar != '\n');
      break;
    case 'p':
      in >> buf >> itsNumNodes >> itsNumEdges >> itsNumCommodities;
      InitNodes();
      InitCommodities();
      break;

    case 'a':
      in >> u >> v >> val;
      NU = itsNodes[u-1];
      NV = itsNodes[v-1];
	
      NU -> AddEdge(NV);
      NV -> AddEnteringEdge(NU->GetEdge(NV));
      NU -> GetEdge(NV) -> SetCapacity(val*EdgeScale);

      break;
  
    case 'k':
      in >> u >> v >> val;
      itsCommodities[count] = new CCommodity(count, itsNodes[u-1], 
					    itsNodes[v-1], val*CommodityScale);
      count++;
      break;
    case 'e':
      while ((commandChar = getchar()) != EOF && commandChar != '\n');
      break;
    }
  }
}




/*
 * Prepare a graph by creating all the edge queues and 
 * setting up the original values on the sources and sinks. 
 * Don't forget that some node may not have any outgoing or
 * incoming edges.
 */
void CGraph::PrepareGraph(void)  {
  int i;
  CEdge *theEdge;
  
  for (i=0; i<itsNumNodes;i++)  {
    itsNodes[i]->InitEdgeQueues(itsNumCommodities, itsCommodities);
  }

  for (i=0; i < itsNumCommodities; i++)  {
    double originalDemand = itsCommodities[i] -> OriginalDemand();

    theEdge = (itsCommodities[i]->Source()) -> FindSmallestQueueForOutEdge(i);
    if (theEdge != NULL) 
      theEdge -> AdjustQueueSize(tail, i, originalDemand);
    else {
      theEdge = (itsCommodities[i]->Source()) -> FindSmallestQueueForInEdge(i);
      theEdge -> AdjustQueueSize(head, i, originalDemand);
    }

    theEdge = (itsCommodities[i]->Sink()) -> FindLargestQueueForInEdge(i);
    if (theEdge != NULL)  {
      theEdge -> AdjustQueueSize(head, i, -originalDemand);
    } else {
      theEdge = (itsCommodities[i]->Sink()) -> FindLargestQueueForOutEdge(i);
      theEdge -> AdjustQueueSize(head, i, -originalDemand);
    }
  }
}


/*
 * Call this after each iteration.  It will increment the sources
 * and sinks by sigma * original demand, and also change the demands
 * for the commodities (I don't think this is needed outside of the
 * proof though).  Again, some nodes may no have any incoming or
 * outgoing edges.
 */
void CGraph::AdjustSourcesAndSinks(void)  {
  int i;
  double sigma = GetSigma();
  CEdge *theEdge;
  for (i=0; i < itsNumCommodities; i++)  {
    double originalDemand = itsCommodities[i] -> OriginalDemand();
   
    theEdge = (itsCommodities[i]->Source()) -> FindSmallestQueueForOutEdge(i);
    if (theEdge != NULL) 
      theEdge -> AdjustQueueSize(tail, i, sigma * originalDemand);
    else {
      theEdge = (itsCommodities[i]->Source()) -> FindSmallestQueueForInEdge(i);
      theEdge -> AdjustQueueSize(head, i, sigma * originalDemand);
    }
    
    theEdge = (itsCommodities[i]->Sink()) -> FindLargestQueueForInEdge(i);
    if (theEdge != NULL)  {
      theEdge -> AdjustQueueSize(head, i, sigma * -originalDemand);
    } else {
      theEdge = (itsCommodities[i]->Sink()) -> FindLargestQueueForOutEdge(i);
      theEdge -> AdjustQueueSize(tail, i, sigma * -originalDemand);
    }

    itsCommodities[i] -> IncreaseDemand();
  }
}

/*
 * Increment all edge capacities by a sigma fraction of their 
 * original capacities.
 */
void CGraph::AdjustEdgeCapacities(void)  {
  int i;
  for (i=0; i < itsNumNodes; i++)  
    itsNodes[i] -> AdjustEdgeCapacities();
}


/*
 * Cut all flow, queue heights, and capacities in half.
 */
void CGraph::HalveAll(void)  {
  int i;
  for (i=0; i< itsNumCommodities; i++)  
    itsCommodities[i] -> HalveDemand();
  for (i=0; i < itsNumNodes; i++)   {
    itsNodes[i] -> HalveEdgeCapacitiesAndQueues();
  }
}

/*
 * Cut all flow, queue heights, and capacities by the given amount.
 * IE, pass in 2 to divide everything in half.
 */
void CGraph::DivideAll(double amount)  {
  int i;
  for (i=0; i< itsNumCommodities; i++)  
    itsCommodities[i] -> DivideDemand(amount);
  for (i=0; i < itsNumNodes; i++)   {
    itsNodes[i] -> DivideEdgeCapacitiesAndQueues(amount);
  }
}


/*
 * return the amount of the commodity in the graph, taking the
 * on positive values of queue heights.
 */
double CGraph::PositiveAmountOfCommodityInGraph(int commodityIndex)  {
  double amountOfCommodity = 0;
  double val;
  int i;
  for (i = 0; i < itsNumNodes; i++)  {
    val = itsNodes[i] -> 
                ComputePositiveAmountOfCommodity(commodityIndex);  
    amountOfCommodity += val;
  }
  return(amountOfCommodity);
}

/*
 * Compute the total amount of a commodity in the graph.
 * This counts negatives as well.
 */
double CGraph::NetAmountOfCommodityInGraph(int commodityIndex)  {
  double amountOfCommodity = 0;
  double val;
  int i;
  for (i = 0; i < itsNumNodes; i++)  {
    val = itsNodes[i] -> ComputeAmountOfCommodity(commodityIndex);  
    amountOfCommodity += val;
  }
  return(amountOfCommodity);
}



/*
 * Returns TRUE if the excesses in the graph are all at an 
 * acceptable limit.  If this return TRUE, its time to stop.
 * Also, this function will print out information about the
 * upper limit on queue sizes and the acceptable levels.
 */
int CGraph::ExcessesWithinBounds(void)  {
  int i;
  double amountOfCommodity;
  int flag = TRUE;
  for (i=0; i< itsNumCommodities; i++)  {
    int ok;
    double demand = itsCommodities[i] -> OriginalDemand();
    ok = (PositiveAmountOfCommodityInGraph(i) < GetEpsilon() * demand);
    printf("    %4d        %c       %-8.4g      %-8.4g  \n", 
	   i, (ok ? ' ' : 'X'), PositiveAmountOfCommodityInGraph(i), 
	   GetEpsilon() * demand);
    
    flag = flag && ok;
  }
  return (flag);
}

/*
 * Returns TRUE if the excesses in the graph are all at an 
 * acceptable limit.  If this return TRUE, its time to stop.
 * Also, this function will print out information about the
 * upper limit on queue sizes and the acceptable levels.
 */
int CGraph::NumExcessesWithinBounds(void)  {
  int i;
  int count = 0;
  double amountOfCommodity;
  int flag = TRUE;
  for (i=0; i< itsNumCommodities; i++)  {
    double demand = itsCommodities[i] -> OriginalDemand();
    if (PositiveAmountOfCommodityInGraph(i) < GetEpsilon() * demand)
      count++;
  }
  return (count);
}


/*
 * 
 */
void CGraph::BalanceEdges(void)  {
  int i;
  for (i = 0; i < itsNumNodes; i++)  
    itsNodes[i] -> BalanceAcrossEdges();
}	

/*
 * 
 */
void CGraph::BalanceNodes(void)  {
  int i;
  for (i = 0; i < itsNumNodes; i++)  
    itsNodes[i] -> BalanceWithinNode();
}	


/*
 * The following three functions just print random table headings and
 * such.
 */
void CGraph::PrintHeader(void)  {
  cout << "\n\n******************************************"
          "*****************************\n\n";
  cout << "Parameters:\n";
  cout << "-----------\n";
  PrintParameterFunctions();
}

void CGraph::PrintStatus(int round, int phase, double origExcess)  {
  cout << "\n\nPhase: " << phase << "  (" << round << " rounds)";
  cout << "\n------\n";
  cout << "    epsilon = " << GetEpsilon()  << "\n";
  cout << "    sigma   = " << GetSigma() << "\n";
  cout << "    alpha   = " << GetAlpha() << "\n";
  cout << "    Excess = " << ComputeAmountOfPositiveExcess() <<
		" of " << origExcess << " (" <<
		ComputeAmountOfPositiveExcess() / origExcess * 100 << "%)\n";
  cout << "   Commodity    OK?     Excess    Max OK level  \n";
  cout << "  ---------------------------------------------\n";
}

void CGraph::PrintFlow(void)  {
  int i;
  cout << "\n\n\n";
  cout << "Flow:\n-----\n";
  cout << "      ";
  cout << "\n";
  for (i=0;i<itsNumNodes;i++) {
    itsNodes[i]->PrintFlows();
  }
  cout << "\n\n";
}

void CGraph::GiveUp(double originalExcess)  {
  cout << "*******************************************************\n";
  cout << "* Demands cannot be satisfied.  Best solution had \n";
  cout << "* " <<  ComputeAmountOfPositiveExcess() << " of " <<
    originalExcess << " (" <<  
      ComputeAmountOfPositiveExcess() /originalExcess * 100 <<
	"%) excess left. \n\n\n";
  cout << "*******************************************************\n";
}


/*
 * main control loop.  Iterates until every commodity in the graph
 * has an excess of less than epsilon percent of the oroginal demand.
 * This will recognize an impossible demand by halting as soon as 
 * the graph's total excess level does not drop by more than 3 % in
 * one phase.
 */
void CGraph::Balance(void)  {
  int rounds = 0;                    /* number of iterations so far */
  int phase = 0;                     /* number of phases */
  int counter = 0;                   /* when this equals 1/sigma, new phase */
  double sigma;                
  double originalExcess, lastExcess; /* used to determine impossible graph */
  int divisor = 2;                   /* phases last divisor/sigma iterations */

#ifdef BOOST_DEGREE
  BoostDegree();
#endif

  /* init everything */
  PrepareGraph();
  lastExcess = originalExcess = ComputeAmountOfPositiveExcess();
  ComputeParameters(this);
  PrintHeader();
  PrintStatus(rounds, phase, originalExcess);
  ExcessesWithinBounds();

  sigma = GetSigma();

  while (TRUE)  {
    
    /* part 1 */
    AdjustSourcesAndSinks();

    /* part 2 */
    BalanceEdges();
    
    /* part 3 */
    BalanceNodes();
    
    if (counter >= (divisor - 1) / sigma)   {
      /* adjust counters, demands, parameters */
      phase++;
      DivideAll(divisor);
      ComputeParameters(this);
      sigma = GetSigma();
      PrintStatus(rounds, phase, originalExcess);

      /* time to quit if it is */
      if (ExcessesWithinBounds()) 
	break;

      /* only check after first few phases since the excess may increase
         right at the start */
      if (phase > 2 && lastExcess * (1 - kMinDecreasePerPhase) < 
	  ComputeAmountOfPositiveExcess())  {
	GiveUp(originalExcess);
	break;
      }
      
      lastExcess = ComputeAmountOfPositiveExcess();


      counter = 0;
    } else {
      /* keep going in same phase */
      AdjustEdgeCapacities();
      counter++;
    }		
    rounds++; 
  }
}

double CGraph::ComputeAmountOfPositiveExcess(void)  {
  double amount = 0.0;
  int i;
  for (i=0;i<itsNumCommodities;i++)  {
    amount += PositiveAmountOfCommodityInGraph(i);
  }
  return amount;
}


/*
 * The checks for NULL are needed in the case of undirected graphs only.
 */
double CGraph::GetLargestExcess(void)  {
  int i,j;
  double max = 0;
  CEdge *edge;
  int com;
  double t;
  for (i=0; i < itsNumCommodities; i++)  
    for (j = 0; j < itsNumNodes; j++)  {
      edge = itsNodes[j] -> FindLargestQueueForOutEdge(i);
      if (edge!=NULL)  {
	t = edge->GetQueueSize(tail,i) / itsCommodities[i]->OriginalDemand();
	if (max < t)  {
	  max = t;
	}
      }
      edge = itsNodes[j] -> FindLargestQueueForInEdge(i);
      if (edge != NULL)  {
	t = edge->GetQueueSize(tail, i) / itsCommodities[i]->OriginalDemand();
	if (max < t)  {
	  max = t;
	}
      }
    }
  return (max);
}

void CGraph::BoostDegree(void)  {
  int averageDegree = (2 * itsNumEdges) / itsNumNodes;
  int i;

  for (i=0;i<itsNumNodes;i++) {
    int degree;
    for (degree = itsNodes[i] -> GetDegree(); degree < averageDegree; 
	 degree+=2)  {
      cout << "Adding dummy edge (" << i << "  -> " << i << ")\n";
      itsNumEdges++;
      itsNodes[i] -> AddEdge(itsNodes[i]);
      itsNodes[i] -> AddEnteringEdge(itsNodes[i]->GetEdge(itsNodes[i]));
    }	   
  }
}









