/******************************************************************************

  CGraph.h

  Stephen Freund
  freunds@cs.stanford.edu

  The CGraph class manages everything.  It reads the graph data in from
  a stream, sets up the data structures for the graph, and will compute
  the optimal flow using the balancing algorithm.  The Balance method
  will perform the actual balancing.

  There are some switches in Parameter.h that determine the format 
  of the graph (directed/undirected, etc.).

  This object has methods to adjust all elements of the graph, including
  demands, capacities, excess levels, and parameters.  These are used in
  the Balance method to run the computation.  For operations on the whole
  graph, the CGraph object will map an operation to all the nodes, which 
  will then either perform the actions or make all exiting edges to the
  action.  

  Don't forget to set the input type to what you want to use.

******************************************************************************/




#ifndef _H_CGraph
#define	_H_CGraph

#include <iostream.h>

/* 
 * capacity of each edge in the graph.
 */
#define CAPACITY			1500

class CCommodity;
class CNode;

class CGraph {

  private:

    int itsNumEdges;              /* number of edges in the graph */
    int itsNumNodes;              /* number of nodes */
    CNode **itsNodes;             /* array of node objects */
    int itsNumCommodities;        /* number of commodities */
    CCommodity **itsCommodities;  /* array of commodity objects */

    /* random helper functions to initialize, read, and print 
       various things */
    void InitNodes(void);
    void InitCommodities(void);
    void ReadTopology(istream& in, double scale = 1);
    void ReadCommodities(istream& in, double scale = 1);
    void PrintHeader(void);
    void PrintStatus(int round, int phase, double origExcess);
    void PrepareGraph(void);
    void BoostDegree(void);
    void GiveUp(double originalExcess);

  public:
    /* constructors / destructor */
    CGraph();
    CGraph(int numberOfNodes);
    
    /* Accessors */
    int GetNumEdges(void) const        { return itsNumEdges; }
    int GetNumNodes(void) const	       { return itsNumNodes; }
    int GetNumCommodities(void) const  { return itsNumCommodities; }
    CNode *GetNode(int index) const   { return itsNodes[index]; }
    CCommodity *GetCommodity(int index) const { return itsCommodities[index]; }
    double GetLargestExcess(void);
    double ComputeAmountOfPositiveExcess(void);
    
    /* Adjusting levels of the queues and flows */
    void AdjustSourcesAndSinks(void);
    void AdjustEdgeCapacities(void);
    void HalveEdgeCapacities(void);
    
    /* balancing stuff */
    void HalveAll(void);
    void DivideAll(double amount);
    double PositiveAmountOfCommodityInGraph(int commodityIndex);
    double NetAmountOfCommodityInGraph(int commodityIndex);
    int ExcessesWithinBounds(void);
    int NumExcessesWithinBounds(void);
    void BalanceEdges(void);
    void BalanceNodes(void);		

    /* This function does all the work */
    void Balance(void);

    /* I/O */
    friend ostream& operator<<(ostream&,const CGraph&);
    friend istream& operator>>(istream&, CGraph&); /* same as matrix */
    void ReadGraphFromMatrixInput(istream& in, double EdgeScale = 1, 
				  double CommodityScale = 1);
    void ReadGraphFromFormattedInput(istream& in, double EdgeScale = 1, 
				  double CommodityScale = 1);
    void PrintFlow(void);
  };
   

#endif








